#charset "us-ascii"

/*
 *   Copyright 2003-04 Soren J. Lovborg.
 *  
 *   Permission is hereby granted, free of charge, to any person obtaining a
 *   copy of this software and associated documentation files (the "Software"), 
 *   to deal in the Software without restriction, including without limitation 
 *   the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *   and/or sell copies of the Software, and to permit persons to whom the
 *   Software is furnished to do so, subject to the following conditions:
 *  
 *   The above copyright notice and this permission notice shall be included in
 *   all copies or substantial portions of the Software.
 *  
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *   DEALINGS IN THE SOFTWARE.
 *
 */

#include <tads.h>
#include <tok.h>
#define false nil

#define OPTIMIZE

/* ------------------------------------------------------------------------ */

class CodeFragment: object
    code = nil
    construct(code_) { code = code_; }
    
    stringValue(nextCode)
    {
        local x = code.prepareCode();
#ifdef OPTIMIZE
        if (x) x = x.optimize();
#endif
        return x ? x.programCode() + '\n' : '';
    }
;

class CodeObject: object
    fields = nil
    id = nil
    type = nil
    flags = nil
    
    getConstValue(x, errWhich)
    {
        x = x.prepareCode();
        if (!x || !x.ofKind(CompilerConstant)) throw new ECompilerError(errWhich + ' must be a constant');
        return x.programCode();
    }
        
    construct(identifierToken)
    {
        local jcc = jccCompiler;

        fields = new Vector(8);
        flags = new Vector(2);
                
        type = jcc.enforceToken(ObjectTypeToken).type;
        id   = jcc.rParenthesis.constantNumber;
        if (id == nil) throw new ECompilerError('Object ID must be a constant number');

        jcc.currentlyDefiningObject = self;
        if (identifierToken) identifierToken.assignValue(new CompilerConstantNumber(id));

        if (jcc.nextToken is in (rwIn, rwOn))
        {
            jcc.readToken();
            fields.append('location');
            fields.append(jcc.rParenthesis);
        }
        
        if (jcc.nextToken == opHas)
        {
            jcc.readToken();
            while (jcc.nextToken != tBraceBegin)
            {
                if (flags.length() != 0) jcc.enforceToken(tComma);
                flags.append(jcc.rExpression);
            }
        }
                
        jcc.enforceToken(tBraceBegin);
                
        while (jcc.nextToken != tBraceEnd)
        {
            local params = new Vector(3);
            local fieldname;
            
            if (jcc.nextToken.ofKind(StringToken)) fieldname = jcc.readToken().value();
            else fieldname = jcc.enforceToken(WordToken).token;

            if (jcc.nextToken == tBracketBegin)
            {
                jcc.readToken();
                jcc.enforceToken(tBracketEnd);

                if (jcc.nextToken is in (tColon, tAssign)) jcc.readToken();

                jcc.enforceToken(tBracketBegin);

                local i = 0;
                while (jcc.nextToken != tBracketEnd)
                {
                    if (i != 0) jcc.enforceToken(tComma);

                    fields.append(fieldname + i);
                    fields.append(jcc.rExpression);

                    i++;
                }
                jcc.arrayLengths[[id, fieldname]] = i;

                jcc.readToken(); // Skip the ']'.    
                jcc.enforceToken(tSemicolon);
                continue;
            }
            
            fields.append(fieldname);

            if (jcc.nextToken == tParenBegin)
            {
                jcc.readToken();

                while (jcc.nextToken != tParenEnd)
                {
                    if (params.length() != 0) jcc.enforceToken(tComma);
                    if (jcc.nextToken.ofKind(WordToken))
                        params.append(new CompilerConstantString(jcc.readToken.token));
                    else params.append(jcc.rExpression);
                }
                jcc.readToken(); // Skip ')'.
            }

            /* Field names and values may be separated by : or = */
            if (jcc.nextToken is in (tColon, tAssign)) jcc.readToken();
                    
            if (params.length() != 0)
            {
                local res = new CompilerPrintCall(); 
                for (local i = 1; i <= params.length(); i++)
                {
                    res.locals.append(params[i]);
                    res.locals.append(
                        new CompilerFieldValue(new CompilerConstantString('%!'), new CompilerConstantString('_cp' + i))
                    );
                }
                        
                res.add(jcc.rExpression);
                fields.append(res);
            }
            else fields.append(jcc.rExpression);
                    
            jcc.enforceToken(tSemicolon);
        }

        jcc.readToken(); // Skip the '}';

        jcc.enforceToken(tSemicolon);
        jcc.currentlyDefiningObject = nil;
    }

    useFieldString(field, default_)
    {
        for (local i = 1; i <= fields.length(); i += 2)
           if (fields[i] == field)
           {
               local x = fields[i + 1];
               fields.removeRange(i, i + 1);
               return getConstValue(x, 'Field "' + field + '"');
           }
        if (!default_) throw new ECompilerError('Object #' + id + ' does not define field "' + field + '"');
        return default_;
    }

    getCreateCommand()
    {
        switch (type)
        {
        case 4: return '@create ' + useFieldString('name', nil);
        case 3: return '@action ' + useFieldString('name', nil) + '=#' + useFieldString('location', nil) + ':#' + useFieldString('action', '-3');
        case 2: return '@pcreate '+ useFieldString('name', nil) + '= (password)';
        case 1: return '@dig '    + useFieldString('name', nil);
        }
        return '(Can\'t create object #' + id + ' of type ' + type + '.)';
    }
    
    stringValue(nextCode)
    {
        local str = getCreateCommand() + ' [as #' + id + ']';
        foreach (local flag in flags) { str += '\n@set #' + id + '=' + getConstValue(flag, 'Flag'); }
        
        for (local i = 1; i <= fields.length(); i += 2)
        {
            str += '\n';
            
            local field = fields[i];

            switch (field)
            {
            case 'description': str += '@desc #' + id + '='; break;
            case 'success':
            case 'osuccess':
            case 'fail':
            case 'ofail': str += '@' + field + ' #' + id + '='; break;
            
            case 'zone': str += '@zone #' + id + '=#' + getConstValue(fields[i+1], 'Zone'); continue;
                
            case 'location': str += '@teleport #' + id + '=#' + getConstValue(fields[i+1], 'Location'); continue;
            default: str += '@field #' + id + '=' + field + ':';
            }

            field = fields[i + 1].prepareCode;
#ifdef OPTIMIZE
            if (field) field = field.optimize();
#endif
            if (field) str += field.programCode();
        }
        return str + '\b';
    }
;

class CodeEmit: object
    str = nil
    construct(str_) { str = str_; }

    stringValue(nextCode)
    {
        return str + (!nextCode || nextCode.ofKind(CodeEmit) ? '\n' : '\b');
    }
;

/* ------------------------------------------------------------------------ */

class CompilerValue: object
    readCode()
    {
        throw new ECompilerError('Attempt to read non-readable value');
    }
    writeCode(newValue)
    {
        throw new ECompilerError('Attempt to assign to non-writable value');
    }
    programCode()
    {
        return readCode();
    }

    construct(readCode_)
    {
        readCode = readCode_;
    }

    optimizeOrPrepare(prop)
    {
        return self;
    }    
    prepareCode() { return optimizeOrPrepare(&prepareCode); }
    optimize()    { return optimizeOrPrepare(&optimize);    }
    
    constantString = nil
    constantNumber = nil
;

class CompilerValueList: CompilerValue
    values = nil
    mayFlatten = false

    add(code) { if (code) values.append(code); }

    prepareCode()
    {
        if (mayFlatten && values.length == 1) return values[1].prepareCode();

        local x = values;
        values = new Vector(x.length());
        for (local i = 1; i <= x.length(); i++)
        {
            local val = x[i].prepareCode();
            if (val) values.append(val);
            else values.append(new CompilerNullValue());
        }
        return inherited;
    }

    optimize()
    {
        values = values.mapAll({o: o.optimize()});
        return inherited;
    }
;

class CompilerJotaCodeList: CompilerValueList
    construct(mayFlatten_)
    {
        values = new Vector(4);
        mayFlatten = mayFlatten_;
    }
    
    readCode()
    {
        local str = '{';
        for (local i = 1; i <= values.length(); i++)
        {
            if (i != 1) str += ',';
            str += values[i].readCode;
        }
        return str + '}';
    }
;

class CompilerConstant: CompilerValue
    createFromString(str)
    {
        if (str == '') return new CompilerNullValue();

        local x = toInteger(str);
        if (str == toString(x)) return new CompilerConstantNumber(x);
        return new CompilerConstantString(str);
    }
;

class CompilerConstantString: CompilerConstant
    construct(str)
    {
        constantString = str;
        if (!str.find('%')) constantNumber = toInteger(str);
        if (str.substr(1, 1) not in (' ', '@')) programCode = str;
    }
    readCode = ('"' + escapeString(constantString) + '"')
    programCode = ('@print(' + readCode + ')')
;

class CompilerConstantNumber: CompilerConstant
    construct(number) { constantString = readCode = toString(constantNumber = number); }
;

class CompilerNullValue: CompilerConstant
    construct() { }
    readCode = '""'
    programCode = ''
    constantString = ''
    constantNumber = 0
;

class CompilerIdentifier: CompilerValue
    token = nil
    construct(token_) { token = token_; }
    prepareCode()
    {
        return token.identifierValue.prepareCode();
    }
;

class CompilerVarObjReference: CompilerValue
    construct() { }
    readCode = '@strcheck(@getfield("%!", "varobj"), "%!")'
;

class CompilerFieldValue: CompilerValue
    field = nil
    obj = nil   
    construct(obj_, field_) { obj = obj_; field = field_; }

    readCode()
    {
        return (obj.ofKind(CompilerVarObjReference) ? '@g(' : '@getfield(' + obj.readCode() + ',')
            + field.readCode() + ')';
    }
    writeCode(newValue)
    {
        return (obj.ofKind(CompilerVarObjReference) ? '@s(' : '@setfield(' + obj.readCode() + ',')
            + field.readCode() + ',' + newValue.readCode() + ')';
    }
    
    optimizeOrPrepare(prop)
    {
        obj = obj.(prop)();
        if (!obj) throw new ECompilerError('Null-object in field-reference');

        field = field.(prop)();
        if (!field) throw new ECompilerError('Null-field in field-reference');

        return self;
    }
;

class CompilerAssignment: CompilerValue
    assignee = nil
    value = nil
    construct(assignee_, value_) { assignee = assignee_; value = value_; }
    readCode() { return assignee.writeCode(value); }

    optimizeOrPrepare(prop)
    {
        assignee = assignee.(prop)();
        if (!assignee) assignee = new CompilerNullValue();

        value    = value.(prop)();
        if (!value) value = new CompilerNullValue();
        return self;
    }
;

class CompilerMethodCall: CompilerValueList
    callee = nil
    construct(callee_) { callee = callee_; values = new Vector(1); }

    prepareCode()
    {
        inherited();
        
        callee = callee.prepareCode();
        if (!callee || !callee.ofKind(CompilerFieldValue))
            throw new ECompilerError('Cannot call non-field value');
        
        local callobj = callee.obj;
        local callfield = callee.field;
        local x, res;

        if (values.length() != 0)
        {
            res = new CompilerPrintCall();

            local callobjstr = callobj.constantString();
            if (!callobjstr)
            {
                res.locals.append(new CompilerConstantString('_co_'));
                res.locals.append(callobj);
                callobjstr = '%{_co_}';
            }
            callobj = CompilerConstant.createFromString(callobjstr);

            for (local i = 1; i <= values.length(); i++)
            {
                local str = values[i].constantString();
                if (!str)
                {
                    res.locals.append(new CompilerConstantString('_cp' + i));
                    res.locals.append(values[i]);
                    str = '%{_cp' + i + '}';
                }
                values[i] = CompilerConstant.createFromString(str);
            }

            for (local i = 1; i <= values.length(); i++)
            {
                x = new CompilerFunctionCall('@setfield');
                x.add(CompilerConstant.createFromString(callobjstr)); 
                x.add(new CompilerConstantString('_cp' + i));
                x.add(values[i]); 
                res.add(x);
            }
        }

        x = new CompilerFunctionCall('@call');
        x.add(callobj);
        x.add(callfield);
        if (res) { res.add(x); return res.prepareCode(); }
        return x.prepareCode();
    }
;

class CompilerFunctionCall: CompilerValueList
    construct(funcName_)
    {
        funcName = funcName_;
        values = new Vector(3);
    }
    funcName = nil
    
    readCode()
    {
        local str = funcName + '(';
        for (local i = 1; i <= values.length(); i++)
        {
            if (i != 1) str += ',';
            str += values[i].readCode;
        }
        return str + ')';
    }
;

class CompilerPrintCall: CompilerOperator
    construct() { inherited(opConcat); locals = new Vector(2); }

    mayFlatten = (locals.length() == 0)

    locals = nil

    prepareCode()
    {
        local x = inherited();
        
        if (locals.length() >= 1)
        {
            local let = new CompilerFunctionCall('@let');
            for (local i = 1; i <= locals.length(); i++)
                let.add(locals[i].prepareCode());

            let.add(x);
            locals.setLength(0);
            return let.prepareCode();
        }
        
        return x;
    }

    constantString()
    {
        local str = '';
        for (local i = 1; i <= values.length(); i++)
        {
            local x = values[i].constantString();
            if (x == nil) return nil;
            str += x;
        }
        return str;
    }
    
    optimize()
    {
        inherited();
        /* We shouldn't optimize locals, as they've already been made into a separate @let-function */

        if (values.length() == 0) return new CompilerNullValue();

        local values_ = new Vector(values.length());
        for (local i = 1; i <= values.length(); i++)
        {
            if (values[i].ofKind(CompilerPrintCall) && values[i].locals.length() == 0)
                values_.appendAll(values[i].values);
            else values_.append(values[i]);
        }

        local prevval = values_[1].constantString();
        for (local i = 2; i <= values_.length(); i++)
        {
            local curval = values_[i].constantString();
            if (prevval != nil && curval != nil)
            {
                values_.removeElementAt(i); /* Remove curval */
                i--;
                prevval = prevval + curval; /* Concat prevval & curval */
                values_[i] = CompilerConstant.createFromString(prevval); /* And put in prevval's place */
                continue;
            }
            prevval = curval;
        }

        values = values_;
        return values.length() == 1 ? values[1] : self;
    }
;

/* ------------------------------------------------------------------------ */

jccCompiler: object
    tokens = nil

    currentlyDefiningObject = nil
    arrayLengths = static new LookupTable(4, 8)
    
    compile()
    {
        nextTokenIndex = 0;
        nextToken = nil;
        readToken();

        currentlyDefiningObject = nil;

        local code = new Vector(16);

        while (nextToken != eofToken)
        {
            if (nextToken == rwConst)
            {
                readToken();
                do
                {
                    local tok = enforceToken(IdentifierToken);
                    enforceToken(tAssign);
                    tok.assignValue(rExpression);
                } while (nextToken == tComma && readToken());
                enforceToken(tSemicolon);
            }
            else if (nextToken.ofKind(EmitToken))
            {
                code.append(new CodeEmit(readToken().value()));
            }
            else if (nextToken.ofKind(ObjectTypeToken) && peekAheadToken != tParenBegin)
            {
                code.append(new CodeObject(nil));
            }
            else if (nextToken.ofKind(IdentifierToken) && peekAheadToken == tColon)
            {
                local x = readToken();
                readToken(); // Skip ':'.
                code.append(new CodeObject(x));
            }
            else
            {
                code.append(new CodeFragment(rProgram()));
            }
        }

        nextTokenIndex = nil;
        
        local str = '';
        for (local i = 1; i <= code.length(); i++)
            str += code[i].stringValue(i < code.length() ? code[i + 1] : nil);
        return str;
    }
    
    nextToken = nil
    nextTokenIndex = nil
    peekAheadToken = nil  /// DANGEROUS HACKY VARIABLE - DO NOT USE UNLESS YOU REALLY, REALLY HAVE TO! ///
    
    requestMoreTokens = nil
    eofIsValid = false
    
    readToken()
    {
        local x = nextToken;

        nextToken = ++nextTokenIndex <= tokens.length() ? tokens[nextTokenIndex] : eofToken;

        if (nextToken == eofToken && requestMoreTokens)
        {
            requestMoreTokens()();
            nextToken = nextTokenIndex <= tokens.length() ? tokens[nextTokenIndex] : eofToken;
        }
        peekAheadToken = nextTokenIndex < tokens.length() ? tokens[nextTokenIndex + 1] : eofToken;
       
        return x;
    }
    enforceToken(toktype)
    {
        local x = readToken();

        if (!x.ofKind(toktype)) throw new EInvalidToken(toktype.className, x);
        return x;
    }
    
    rProgram()
    {
        local res = new CompilerPrintCall();

        while (!(nextToken == eofToken || nextToken.ofKind(EmitToken) ||
            (nextToken.ofKind(ObjectTypeToken) && peekAheadToken != tParenBegin) ||
            (nextToken.ofKind(IdentifierToken) && peekAheadToken == tColon)))
        {
            local x = rStatement();
            enforceToken(tSemicolon);
            if (x) res.add(x);
        }
        return res;
    }
    rStatement()
    {
        local res;
        
        if (nextToken == tSemicolon) return nil; // empty statement.

        res = rExpression();
        if (nextToken == tAssign)
        {
            readToken();
            res = new CompilerAssignment(res, rExpression);
        }
        return res;
    }

    readAssociative(wantType, subCall, allowMultiRepetitions)
    {
        local res = self.(subCall);
        local result;
        
        while (wantType.indexOf(nextToken))
        {
            local tok = readToken();

            if (!result) { result = tok.getCompilerValue(); result.add(res); }
            result = result.addPart(tok, self.(subCall));
            if (!allowMultiRepetitions) break;
        }
        return result ? result : res;
    }

    rExpression()
    {
        local res = rConcat();
        if (nextToken == tConditional)
        {
            readToken();

            local x = res;
            res = new CompilerFunctionCall('@switch');
            res.add(new CompilerConstantNumber(1));
            res.add(x);
            res.add(rExpression);
            enforceToken(tColon);
            res.add(rExpression);
        }
        return res;
    }

    rConcat()    { return readAssociative([opConcat],  &rBoolAndOr, true); }    
    rBoolAndOr() { return readAssociative([opAnd, opOr], &rBoolXor, true); }
    rBoolXor()   { return readAssociative([opXor],    &rBoolNegate, true); }
    rBoolNegate()
    {
        if (nextToken == opNot)
        {
            readToken();
            local x = opNot.getCompilerValue();
            x.add(rBoolNegate);
            return x;
        }
        return rHasHasnt;
    }
    rHasHasnt()   { return readAssociative([opHas, opHasnt], &rIs, false); }
    rIs()
    {
        local res = rComparison();
        if (nextToken not in (rwIs, rwIsnt)) return res;
        local negate = readToken() == rwIsnt;
        
        local a, b;
        
        if (nextToken is in (rwIn, rwOn))
        {
            readToken();
            a = new CompilerFunctionCall('@location');
            b = rMethodCall();
        }
        else
        {
            a = new CompilerFunctionCall('@type');
            b = new CompilerConstantNumber(enforceToken(ObjectTypeToken).type);
        }

        a.add(res);
        res = opEqual.getCompilerValue();
        res.add(a);
        res.add(b);
        
        if (negate) { a = opNot.getCompilerValue(); a.add(res); return a; }
        return res;
    }
    rComparison() { return readAssociative([opNotEqual, opLess, opGreater, opLessEqual, opGreaterEqual], &rEquality, false); }
    rEquality()   { return readAssociative([opEqual], &rSum, true); }
    rSum()        { return readAssociative([opPlus, opMinus], &rProduct, true); }
    rProduct()    { return readAssociative([opMultiply, opDivide, opIDivide, opModulos], &rNegate, true); }
    rNegate()
    {
        if (nextToken == opMinus)
        {
            readToken();
            local x = opNeg.getCompilerValue();
            x.add(rNegate());
            return x;
        }
        return rMethodCall;
    }

    rMethodCall()
    {
        local res = rFieldExpression();
        if (nextToken == tParenBegin)
        {
            readToken();
            res = new CompilerMethodCall(res);
            local requireComma = false;
            while (nextToken != tParenEnd)
            {
                if (requireComma) enforceToken(tComma); else requireComma = true;
                res.add(rExpression);
            }
            readToken(); // Skip the ')'.
        }
        return res;
    }

    rFieldExpression()
    {
        local res;
        if (nextToken == tPeriod) res = new CompilerVarObjReference();
        else res = rParenthesis();

        while (true)
        {
            if (nextToken == tPeriod)
            {
                readToken();

                local field = new CompilerPrintCall(); 
                if (!(nextToken.ofKind(WordToken) || nextToken == tBracketBegin))
                    throw new EInvalidToken('WordToken or tBracketBegin', nextToken);
                
                if (nextToken.ofKind(WordToken))
                {
                    field.add(new CompilerConstantString(readToken.token));
                }
                if (nextToken == tBracketBegin)
                {
                    readToken();
                    field.add(rExpression);
                    enforceToken(tBracketEnd);
                }
                res = new CompilerFieldValue(res, field);
            }
            else return res;
        }
    }

    rParenthesis()
    {
        if (nextToken == tParenBegin) { readToken(); local x = rExpression(); enforceToken(tParenEnd); return x; }
        return rSubExpression;
    }

    rSubExpression()
    {
        local res, tok = readToken(), requireComma, paren, x;
        
        switch (tok)
        {
        case tBraceBegin:
            res = new CompilerPrintCall();
            
            while (nextToken == rwLet)
            {
                readToken();
                res.locals.append(rExpression);
                enforceToken(tAssign);
                res.locals.append(rExpression);
                enforceToken(tSemicolon);
            }

            while (nextToken != tBraceEnd) { res.add(rStatement); enforceToken(tSemicolon); }

            readToken(); // Skip the '}'.
            return res;

        case rwIf:
            res = new CompilerFunctionCall('@switch');
            res.add(new CompilerConstantNumber(1));
            x = (nextToken == tParenBegin);
            if (x) readToken();
            res.add(rExpression);
            if (x)
            {
                enforceToken(tParenEnd);
                if (nextToken == rwThen) readToken();
            }
            else enforceToken(rwThen);
             
            res.add(rStatement);
            
            while (nextToken == tSemicolon && peekAheadToken == rwElse)
            {
                readToken(); readToken();
                if (nextToken == rwIf)
                {
                    readToken();
                    res.add(rExpression);
                    if (nextToken == rwThen) readToken();
                }
                res.add(rStatement);
            }
            return res;
        
        case rwSwitch:
            res = new CompilerFunctionCall('@switch');
            res.add(rExpression);
            enforceToken(tBraceBegin);

            while (nextToken not in (tBraceEnd, rwDefault))
            {
                enforceToken(rwCase);

                x = new CompilerJotaCodeList(true);
                x.add(rExpression);

                while (nextToken == tComma) { readToken(); x.add(rExpression); }
                enforceToken(tColon);
                
                res.add(x);
                
                x = new CompilerPrintCall();                
                while (nextToken not in (tBraceEnd, rwCase, rwDefault))
                {
                    x.add(rStatement);
                    enforceToken(tSemicolon);
                }
                res.add(x);
            }

            if (nextToken == rwDefault)
            {
                readToken(); // Skip the 'default'.
                enforceToken(tColon);
                while (nextToken != tBraceEnd) { res.add(rStatement); enforceToken(tSemicolon); }
            }

            readToken(); // Skip the '}'.
            return res;

        case rwContentsloop:
        case rwStrloop:
            res = new CompilerFunctionCall(tok == rwStrloop ? '@strloop' : '@contentsloop');

            paren = (nextToken == tParenBegin);
            if (paren) readToken();
            tok = rExpression();
            enforceToken(rwIn);
            res.add(rExpression);
            res.add(tok);
            if (paren) enforceToken(tParenEnd);
            res.add(rStatement);
            return res;

        case rwFieldloop:
            res = new CompilerFunctionCall('@fieldloop');
            paren = (nextToken == tParenBegin);
            if (paren) readToken();
            res.add(rExpression);
            enforceToken(rwBy);
            res.add(rExpression);
            if (paren) enforceToken(tParenEnd);
            res.add(rStatement);
            return res;

        case rwRepeat:
            tok = jccTokenizer.identifiers['repeat_string'];
            if (!tok) throw new ECompilerError('repeat needs you to define a repeat_string');
            
            x = new CompilerFunctionCall('@substr');
            x.add(tok.getCompilerValue());
            x.add(new CompilerConstantNumber(0));
            x.add(rExpression);

            res = new CompilerFunctionCall('@strloop');
            res.add(x);
            res.add(new CompilerConstantString('repeat_'));
            res.add(rStatement);
            return res;
        
        case rwSay:
            res = new CompilerFunctionCall('@tell');
            res.add(new CompilerConstantString('%#'));
            res.add(rExpression);
            return res;

        case rwOSay:
        case rwRSay:
            x = new CompilerFunctionCall('@location');
            x.add(new CompilerConstantString('%#'));
            
            res = new CompilerFunctionCall('@tellroom');
            res.add(x);
            res.add(tok == rwOSay ? new CompilerConstantString('%#') : new CompilerNullValue());
            res.add(rExpression);
            return res;

        case tBracketBegin:
            res = new CompilerJotaCodeList(false); // May not flattent one-item lists.
            requireComma = false;
            while (nextToken != tBracketEnd)
            {
                if (requireComma) enforceToken(tComma); else requireComma = true;
                res.add(rExpression);
            }
            readToken(); // Skip the ']'.
            return res;
            
        case rwArrayLen:
            res = rParenthesis();
            if (res.ofKind(CompilerFieldValue))
            {
                x = res.obj;
                res = res.field;
            }
            else x = rwThis.getCompilerValue();
            x = x.constantNumber;
            if (x == nil) throw new ECompilerError('Non-constant object for arraylen keyword');
            res = res.constantString;
            if (res == nil) throw new ECompilerError('Non-constant field for arraylen keyword');

            tok = arrayLengths[[x, res]];
            if (!tok) throw new ECompilerError('The definition of object #' + x + ' did not contain any array "' + res + '"');
            
            return new CompilerConstantNumber(tok);
        }
        
        if (tok.ofKind(ValueToken)) return tok.getCompilerValue();

        if (tok.ofKind(JCFunctionToken))
        {
            res = new CompilerFunctionCall('@' + tok.token);

            requireComma = false;
            
            enforceToken(tParenBegin);
            while (nextToken != tParenEnd)
            {
                if (requireComma) enforceToken(tComma); else requireComma = true;
                res.add(rExpression);
            }
            readToken(); // Skip the ')'.
            return res;
        }

        throw new EInvalidToken('expression', tok);
    }
;

/* ------------------------------------------------------------------------ */
