#charset "us-ascii"

/*
 *   Copyright 2003-04 Soren J. Lovborg.
 *  
 *   Permission is hereby granted, free of charge, to any person obtaining a
 *   copy of this software and associated documentation files (the "Software"), 
 *   to deal in the Software without restriction, including without limitation 
 *   the rights to use, copy, modify, merge, publish, distribute, sublicense,
 *   and/or sell copies of the Software, and to permit persons to whom the
 *   Software is furnished to do so, subject to the following conditions:
 *  
 *   The above copyright notice and this permission notice shall be included in
 *   all copies or substantial portions of the Software.
 *  
 *   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *   FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *   DEALINGS IN THE SOFTWARE.
 *
 */

#include <tads.h>
#include <tok.h>
#include <file.h>
#define false nil

/* ------------------------------------------------------------------------ */
/*
 *  To-do
 *  -----
 *
 *  + #include.
 *  + Optimizations. (Switch/IF)
 */
 
function escapeString(str)
{
    return rexReplace('<backslash|dquote>', str, '\\%*', ReplaceAll);
}


EMessage: Exception
    construct(msg_) { msg = msg_; }
    displayException = "<< msg >>. "
    msg = nil
;

ESourceError: EMessage
    lineNo = nil
    linePos = nil

    displayException()
    {
        if (lineNo != nil) " at << lineNo >>: << linePos >>";
        ": << msg >>. ";
    }
;

ETokenizerError: ESourceError
    construct(msg_)
    {
        inherited(msg_);
        lineNo = jccTokenizer.sourceLineNo;
        linePos = jccTokenizer.sourceLinePos;
    }
    displayException() { "Tokenizer error"; inherited; }
;

ECompilerError: ESourceError
    construct(msg_)
    {
        inherited(msg_);
        try
        {
           lineNo = jccTokenizer.tokenPositions[jccCompiler.nextTokenIndex * 2 - 1];
           linePos = jccTokenizer.tokenPositions[jccCompiler.nextTokenIndex * 2];
        }
        catch (RuntimeError e)
        {
            lineNo = nil;
            linePos = nil;
        }
    }
    displayException() { "Compiler error"; inherited; }
;

ETokenizerReachedEOF: ETokenizerError construct() { inherited('Reached unexpected EOF'); };
ECompilerReachedEOF:  ECompilerError  construct() { inherited('Reached unexpected EOF'); };

EInvalidToken: ECompilerError
    construct(expected, foundTok)
    {
        inherited('Expected ' + expected + ' but found ' + foundTok.className + (foundTok.value ? ' (' + foundTok.value + ')' : ''));
    }
;

globals: object
    file = nil
    outputFile = nil
    
    firstInput = true
    compilerWantsInput = false
    inputReachedEOF = false
    noPrompt = false

    readKeyboardInput()
    {
        local str;
        while (true)
        {
            if (!noPrompt) { if (firstInput) ">\ "; else "&emsp;"; }
        
            str = inputLine();
            if (str == '') continue;

            if (str && str.length() == 2 && str.toUnicode(1) == 35)
            {
                switch (str.substr(2, 1).toLower())
                {
                case 'q': str = nil; break;
                case 'c':
                    "<B>Credits</B>\n
                     Compiler:\tS\u00F8ren J.\ L\u00F8vborg.\n
                     Language:\tDavidW, schep, katre, BrenBarn and the rest of #coding.\n
                     JotaCode:\tJota, DGlasser, inky, thumper, markm, jwalrus, Jaybird, and other ifMUDders.\n
                     TADS 3:\tMike J.\ Roberts\b
                     You can find JCC at <A href='http://kwi.homepage.dk/kwif/jcc.htm'>kwi.homepage.dk/kwif/jcc.htm</A>
                     and ifMUD at <A href='http://ifmud.port4000.com'>ifmud.port4000.com</A>.\b";
//                     [coding] DavidW says, \"Boo!Boo!Boo!\"\b";
                    continue;
                }
            }
            if (str) { firstInput = false; return str + '\n'; }

            inputReachedEOF = true;
            return nil;
        }
    }
;

function fetchSource()
{
    local str;
    if (globals.file) str = globals.file.readFile();
    else if (globals.compilerWantsInput)
    {
        globals.compilerWantsInput = false;
        str = globals.readKeyboardInput();
    }
    if (str == nil) throw new ETokenizerReachedEOF();
    jccTokenizer.source += str;
}

function fetchTokens()
{
    /* El Grande Hack - to allow live-compilation */
    if (!globals.file)
    {
        local hack = t3GetStackTrace()[4];
        if (hack.prop_ == &rProgram) return;
    }
    
    local tokensLen = jccTokenizer.tokens.length();
    if (tokensLen > 0)
    {
       /* Remove the EOF token */
       jccTokenizer.tokens.setLength(tokensLen - 1);
       jccTokenizer.tokenPositions.setLength(jccTokenizer.tokenPositions.length() - 2);
    }
    globals.compilerWantsInput = true;

    jccTokenizer.sourceLinePos = 0;
    jccTokenizer.source = '';
    jccTokenizer.sourceIdx = 0;

    jccTokenizer.tokenize();
}

function doCompile()
{
    try
    {
        jccCompiler.tokens = jccTokenizer.tokens = new Vector(64);
        jccTokenizer.tokenPositions = new Vector(128);

        jccTokenizer.sourceLineNo = 1;

        fetchTokens();
        if (globals.inputReachedEOF) return;

        local res = jccCompiler.compile();
        if (globals.outputFile)
        {
            globals.outputFile.writeFile(res.findReplace('\b', '\n\n', ReplaceAll) + '\n');
        }
        else "\n<< res.htmlify(HtmlifyTranslateWhitespace) >>\b";
    }
    catch (EMessage e)
    {
        "\b<< e.displayException() >>\b";
        if (!(globals.file && e.ofKind(ESourceError) && e.lineNo)) return;

        globals.file.setPos(0);
        for (local i = 2; i < e.lineNo; i++) globals.file.readFile();
        local s1 = globals.file.readFile();
        local s2 = globals.file.readFile();
        
        "<TT><< s1 ? s1.htmlify(HtmlifyTranslateWhitespace) : '' >>";
        if (e.lineNo > 1 && s2) "\n<< s2.htmlify(HtmlifyTranslateWhitespace) >>";

        if (e.linePos > 0)
        {
            "\n";
            if (e.linePos > 1) "<< makeString(0x00a0, e.linePos - 1) >>";
            "^";
        }
        "</TT>\n";
    }
}

function main(args)
{
    "<B>JotaCode Compiler</B> by Kwi Soft, a Loungent Technologies Associate.\b";
     
    globals.noPrompt = (args.length() >= 2 && args[2].toLower() == '-noprompt');
    if (globals.noPrompt) args = args.removeElementAt(2);
    
    if (args.length() >= 3)
    {
        "Too many arguments passed to JCC: Between 0 and 3 required.\n
         Syntax: t3run -plain << args[1] >> [-noprompt] [jc-source] [output-file]\b";
        return;
    }

    jccTokenizer.sourceReadMore = fetchSource;
    jccCompiler.requestMoreTokens = fetchTokens;

    if (args.length() >= 2)
    {
        "Compiling << args[2] >>.\n";
        globals.file = File.openTextFile(args[2], FileAccessRead);

        if (args.length() >= 3)
        {
            globals.outputFile = File.openTextFile(args[3], FileAccessWrite);
        }

        doCompile();

        globals.file.closeFile();
        if (globals.outputFile) globals.outputFile.closeFile();

        "Done, << jccTokenizer.sourceLineNo >> lines.\n";
    }
    else
    {
        if (!globals.noPrompt) "Type #Q to quit and #C for credits.\b";
        while (!globals.inputReachedEOF)
        {
            globals.firstInput = true;
            doCompile();
        }
    }
}

/* ------------------------------------------------------------------------ */
